// 简单的冒泡排序函数的代码和标注
//  要点
//    1 验证函数 bubbleSort 满足给出的冒泡排序算法的函数协议 
//    2 数组一般性质的描述
//    2 采用量化断言描述数组的有序性

#define MAX_LEN 10000

// 函数协议
//     前条件：给出要排序数组 a 的基本性质：
//            数组长度 \length(a)、数据块起始位置 \offset(a) 和数组最大长度限制 MAX_LEN
//     assigns子句：提示数组 a 中的元素（位置）在函数中可能被修改
//     后条件：排序后数组 a 的基本性质没变，数组元素按递增顺序排列
/*@ 
    requires 
        \length(a) == m && \offset(a) == 0 && 0 < m <= MAX_LEN
        ;
    assigns 
        a[0..m-1]
        ;
    ensures  
        \length(a) == m && 0 < m <= MAX_LEN && \offset(a) == 0 
        && (\forall int i : [0..m-2]. a[i] <= a[i+1])
        ;
*/
void bubbleSort(int *const a, int const m) {
    int low, up, j, k, temp;
    low = 0; 
    up = m-1;
    k = up; 
    //    循环不变式
    //        每次排序都会把数组段[low..k]最大的值放在该数组段的最后一位
    //    循环入口（k == up）
    //        尚未排序
    //    循环中 && 循环出口
    //        每次循环后，会选出当前最大元素a[k+1]，将数组分成2部分：
    //        未排序的数组段[low..k]和已排序的数组段[k+1..up-1]。其中
    //        a[k+1]大于[low..k]中的所有元素，同时数组段[k+1..up-1]有序
    /*@ loop invariant 
        (k == up  // 循环入口
         ||
         low <= k < up   // 本层循环的性质
         && (\forall int i:[low..k].a[i]<=a[k+1]) 
         && \forall int i:[k+1..up-1].a[i]<=a[i+1]
        )
        && low == 0 && up == m-1                  
                         // 函数前条件的性质
        && \length(a) == m && \offset(a) == 0 && 0 < m <= MAX_LEN
        ;
    */
    while(k != low) {
        j = low;
        //    循环不变式
        //        在数组段[low..k]内检索，通过交换相邻元素选出最大元素
        //    循环入口（k == up && j == low）
        //        第一次进入循环 k==up && j == low
        //    循环中 && 循环出口
        //        在一轮比较交换过程中，a[j]总是大于数组段a[low..j-1]内的元素
        //        同时具有外层循环的性质
        /*@ loop invariant 
            k != low && low <= j <= k && low == 0 && up == m-1
            && (\forall int i:[low..j-1].a[i]<=a[j]) // 本层循环的性质
            && 
            // 外层循环的性质
            (k == up              
             ||               
             low < k < up 
             && (\forall int i:[low..k].a[i]<=a[k+1]) 
             && (\forall int i:[k+1..up-1].a[i]<=a[i+1])
            )
            // 函数前条件
            && \length(a) == m && \offset(a) == 0 && 0 < m <= MAX_LEN 
            ;
        */
        while( j!=k ) {
            if (a[j] > a[j+1]){
                temp = a[j]; 
                a[j] = a[j+1]; 
                a[j+1] = temp;
            }
            j = j + 1;
        }
        k = k -1;
    }
}


int main(){ // 测试程序
    int arr[] = {3, 2, 1, 4, 6, 7, 5, 8, 10, 9, 20, 18, 19, 16, 15, 13, 14, 17, 12, 11};
    int len = sizeof(arr)/sizeof(arr[0]);

    bubbleSort(arr, len);
    /*@assert  // 确认数组已排序
        \forall int i:[0..len-2].arr[i]<=arr[i+1]
        ;
    */

   return 0;
}